;; guile-present
;; Copyright (C) 2007 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU Lesser General Public License as
;; published by the Free Software Foundation; either version 2.1 of the
;; License, or (at your option) any later version.
;;                                                                  
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
;; Lesser General Public License for more details.
;;                                                                  
;; You should have received a copy of the GNU Lesser General Public
;; License along with this program; if not, contact:
;;
;; Free Software Foundation, Inc.     Voice:  +1-617-542-5942
;; 51 Franklin Street, Fifth Floor    Fax:    +1-617-542-2652
;; Boston, MA  02110-1301,  USA       gnu@gnu.org

;;; Commentary:
;;
;; Utility procedures and macros.
;;
;;; Code:

(define-module (present util)
  :use-module (ice-9 regex)
  :export (match-bind))

(define-macro (define-macro-with-docs name-and-args docs . body)
  `(begin
     (define-macro ,name-and-args ,@body)
     (set-object-property! ,(car name-and-args) 'documentation ,docs)
     (if #f #f)))
(set-object-property! define-macro-with-docs 'documentation
  "Define a macro with documentation.")

(define-macro-with-docs (match-bind regex str vars consequent . alternate)
  "Match a string against a regular expression, binding lexical
variables to the various parts of the match.

@code{match-bind} compiles the regular expression during macro
transformation, so multiple matches will not have to suffer the penalty
of compilation. Because of this, @var{regex} must be a string literal.

@var{vars} is a list of names to which to bind the parts of the match.
The first variable of the list will be bound to the entire match, so the
number of variables needed will be equal to the number of open
parentheses (`(') in the pattern, plus one for the whole match.

@var{consequent} is executed if the given expression @var{str} matches
@var{regex}. If the string does not match, @var{alternate} will be
executed if present. If @var{alternate} is not present, the result of
@code{match-bind} is unspecified.

Here is a short example:
@example
(define (star-indent line)
  \"Returns the number of spaces until the first
   star (`*') in the input, or #f if the first
   non-space character is not a star.\"
  (match-bind \"^( *)\\*.*$\" line (_ spaces)
              (string-length spaces)
              #f))
@end example
"
  (or (string? regex) (error "regex needs to be a string so we can compile it"))
  (let ((re (gensym)) (match (gensym)))
    `(let ((,re ,(make-regexp regex)))
       ((lambda (,match)
          (if ,match
              (apply (lambda ,vars ,consequent)
                     (map (lambda (x) (,match:substring ,match x))
                          (iota (,match:count ,match))))
              ,@alternate))
        (,regexp-exec ,re ,str)))))
